#! /usr/bin/perl

#                   Add Book to Recommendation List

#                   by John Walker  --  October 2001
#                  Revised and updated: September 2004
#                  ISBN-13 support added: October 2005
#             Amazon Query updates, security: September 2019

#   This program is invoked from the CGI script when the
#   user clicks the submit button in the "Recommend Book"
#   form.  Fields in the form are parsed and placed in
#   environment variables as follows.  This is compatible with
#   the "uncgi" program but is, in fact, done internally
#   by the parse_query_string() subroutine in this script.
#
#       WWW_title       Title of book
#       WWW_author      Author's name
#       WWW_isbn        ISBN
#
#   If the user's browser has JavaScript enabled, preliminary
#   validation of these fields will have been performed, but
#   we must not assume that here as a non-JavaScript browser
#   will simply bung in whatever keys the primate pounded on.

    use strict;
    use warnings;
    use Fcntl ':flock';         # Import flock() LOCK_ constants

#   If $email is non-blank, book recommendations will be mailed to
#   the designated address using the $mailer program.  You'll probably
#   want to include a suitable subject with the "-s" option on the
#   $mailer command line.

    my $email = 'mymail@example.com';
    my $mailer = '/bin/mail -s \'Book Recommendation\'';

#   You can override the $email specification given above with a
#   hidden item in the form like the following:
#       <input type="hidden" name="rznvy" value="sbk*zhyqre}sov&tbi">
#   To prevent the E-mail address from being "harvested" and buried
#   in a mountain of spam, it is encoded in the venerable "rot13"
#   coding, which may be performed (coding and decoding are the same
#   operation) with the Unix filter:
#       tr A-Za-z N-ZA-Mn-za-m
#   In addition, you may replace the "@" character in the mail address
#   with either "{" or "}" and periods with any of the characters
#   "?*&" to further disguise the string's identity as an E-mail
#   address.  The field name "rznvy" is the rot13 encoding of
#   "email"; the E-mail address in the example above decodes
#   to "fox.mulder@fbi.gov".

#   If $logfile is non-blank, a CSV-format record containing the
#   recommendation will be appended to the file.  The record is
#   of the form:
#       "ISBN","Author","Title"
#   All fields in this file are quoted whether they need it or
#   not; quotes in strings are represented by two adjacent quote
#   marks.  If the Perl implementation and underlying system
#   supports the flock() function, the log file will be locked
#   while appending the record to avoid race conditions should
#   two copies of this script be executing simultaneously.

    my $logfile = '';

#   You can specify the URL to which the "Back" link points by
#   including a hidden "return" field like the following:
#       <input type="hidden" name="return" value="http://www.myside.dom/whizzer/whizzer.html">
#   If you don't include a "return" item, the Back link will point
#   back to the page which invoked this script, courtesy of
#   the HTTP_REFERER environment item provided by the server.

#   You can specify the text of the link back to the parent
#   document (or whatever document you specified in the "return"
#   item) by including a hidden item of the form:
#       <input type="hidden" name="backlink" value="Back to Fox's UFO Bookshelf">
#   If no "backlink" field appears in the form, the link will
#   bear the imaginative legend "Back".

#   If "logresolve" is installed on your system, specify the path below.
#   IP addresses will be resolved and the domain name displayed if
#   the name resolves.

    my $logresolve = '/usr/bin/logresolve';

#   If $noInvalidISBN is set to 1, all invalid ISBN specifications
#   (the overwhelming majority of which come from form spammers,
#   as the JavaScript catches most legitimate users who use the
#   form) will result in an immediate 500 abort.

    my $noInvalidISBN = 1;

#   Now that spammers attack any form on any system, you can
#   specify a regular expression which, if it matches the ISBN
#   field (which should be well-structured), will abort the submission,
#   rewarding the spammer with a 500 status--not as if that's likely
#   to deter additional attempts.  The important thing is that your
#   mailbox won't be polluted by crap.

    my $forbiddenISBN = qr/(viagra|cialis)/i;

#   And, of course, with the previous protection in effect, we must
#   now filter for spam in the Title field.  Obviously, this should
#   be done with care.  We only apply this test if the ISBN field
#   is blank.

    my $forbiddenTitle = qr/(Verizon|Nokia|ringtones|ringones|check\s+this\+out)/i;

#   Set $debug to include debug information, including a complete dump of
#   the environment variables passed to this process, in the HTML
#   response.

    my $debug = 0;

#   Credit books bought to this Amazon Associate account

    my $assoc = 'myassocacct';

#   You shouldn't need to change anything below this line.

    parse_query_string();

    my $IPaddress = trim($ENV{'REMOTE_ADDR'});
    my $referer;

    if (defined $ENV{'HTTP_REFERER'}) {
        $referer = trim($ENV{'HTTP_REFERER'});
    } else {
        $referer = '';
    }

    #   Process fields from form

    my $title = trim($ENV{'WWW_title'});
    my $author = trim($ENV{'WWW_author'});
    my $sisbn = trim($ENV{'WWW_isbn'});
    my $backlink;
    if (defined $ENV{'WWW_rznvy'}) {        # "rznvy" is rot13("email")
        $email = rot13($ENV{'WWW_rznvy'});  # Decode E-mail alphabetic characters
        $email =~ tr/{}?*&/@@.../;          # Decode alternate stego-delimiters
    }
     if (defined $ENV{'WWW_return'}) {
        $referer = $ENV{'WWW_return'};
    }
     if (defined $ENV{'WWW_backlink'}) {
        $backlink = $ENV{'WWW_backlink'};
    } else {
        $backlink = 'Back';
    }

    #   We have a "subject" field in the form to act as bait for
    #   spammers.  This is a text input field which is hidden
    #   using a CSS style.  Spammers who parse the HTML are likely
    #   to fill in this field with their trash.  If the field is
    #   nonblank, summarily kill the request.

    my $subject = trim($ENV{'WWW_subject'});
    if ($subject ne '') {
        die("Spam inserted in subject field");
    }

    #   Some mindless spam sites directly hit this CGI script
    #   with totally blank $title, $author, and $sisbn fields.
    #   This is usually caught in the JavaScript validation,
    #   but that can't guard against a direct hit here.  If
    #   the recommendation is completely void, ditch it as spam.

    if (($title eq '') && ($author eq '') && ($sisbn eq '')) {
        die("Void recommmendation by spammer");
    }

    #   If this script was invoked from a page lacking the JavaScript
    #   pre-validation of the ISBN (of the the user's browser lacks
    #   JavaScript or has disabled it), we may receive an invalid
    #   ISBN.  Validate the ISBN here; if it fails validation, log
    #   the string the user entered with a flag indicating it is not
    #   a valid ISBN (but may contain a clue as to what the user
    #   intended).

    my $isbn = validate_ISBN($sisbn);
    if (($sisbn ne '') && ($isbn eq '')) {
        if ($noInvalidISBN) {
            die("ISBN Invalid");
        }
        if ($sisbn =~ m/$forbiddenISBN/) {
            die("Forbidden content in ISBN field");
        }
        $isbn = "$sisbn (Incorrect)";
    }

    #   If the ISBN is blank or invalid, apply the spam filter to the
    #   title field and kill the request if it matches.

    if (($isbn eq '') && ($title =~ m/$forbiddenTitle/) &&
        ($author =~ m/^\w+$/)) {
        die("Forbidden content in title field: $title");
    }

    #   Append book recommendation record to the log file.  Ownership
    #   and permissions on the log file need to be such that the userid
    #   under which CGI scripts run may write to it.  If you want the
    #   script to create the log file when the first entry is added,
    #   the directory in which it is placed must also be writable by
    #   the CGI script user.

    my $brec;
    if (($logfile ne '') &&
        ((iquote($isbn) ne '') || ((iquote($author) ne '') && (iquote($title) ne '')))) {
        $brec = $IPaddress;
        $brec .= ' ,' . iquote($isbn);     # Dirty trick: space before comma allows using logresolve
        $brec .= ',' . iquote($author);    # to perform DNS lookup of IP addresses
        $brec .= ',' . iquote($title);
        my @date = localtime(time());
        $brec .= sprintf(',"%04d-%02d-%02d %02d:%02d:%02d"',
                    $date[5] + 1900, $date[4] + 1, $date[3], $date[2], $date[1], $date[0]);
        open(LOGF, ">>$logfile");
        flock(LOGF, LOCK_EX);               # Lock log file
        seek(LOGF, 0, 2);                   # Seek to end in case somebody appended between open and flock
        print(LOGF "$brec\n");              # Append this record to the database
        flock(LOGF, LOCK_UN);               # Unlock the log file
        close(LOGF);                        # Close it
    }

    #   If an E-mail address is specified, send an E-mail with the
    #   recommendation to the designated recipient.

    if (($email ne '') && ($mailer ne '')) {
        open(MAIL, "| $mailer $email");
        print(MAIL "Book recommendation:\n");
        print(MAIL "\n");
        if (($title ne '') || ($author ne '')) {
            print(MAIL "    Title:  $title\n");
            print(MAIL "    Author: $author\n");
            print(MAIL "\n");
            my $eauthor = escape_URL($author);
            my $etitle = escape_URL($title);
            print(MAIL "        https://www.amazon.com/s?i=stripbooks&" .
                "rh=p_27:$eauthor,p_28:$etitle&s=relevanceexprank&tag=$assoc" . "\n");
            print(MAIL "        https://www.abebooks.com/servlet/SearchResults?an=$eauthor&" .
                "bi=0&bx=off&cm_sp=SearchF-_-Advtab1-_-Results&ds=30&recentlyadded=all&" .
                "sortby=17&sts=t&tn=$etitle" . "\n");
        }
        my $eisbn = ISBNtoASIN($isbn);
        if ($eisbn ne '') {
            print(MAIL "\n");
            print(MAIL "    ISBN:   $isbn\n");
            print(MAIL "            US    http://www.amazon.com/dp/$eisbn/?tag=$assoc\n");
            print(MAIL "            UK    http://www.amazon.co.uk/dp/$eisbn/?tag=$assoc\n");
            print(MAIL "            FR    http://www.amazon.fr/dp/$eisbn/?tag=$assoc\n");
            print(MAIL "            DE    http://www.amazon.de/dp/$eisbn/?tag=$assoc\n");
        }
        print(MAIL "\n");
        my $fqdn = '';
        if ($logresolve ne '') {
            $fqdn = `echo $IPaddress | $logresolve`;
        }
        print(MAIL "Submitted by: $IPaddress  $fqdn\n");
        close(MAIL);
    }

    #   Generate confirmation document for submitter.  We show the
    #   submitted ISBN, not the canonicalised version, so the user
    #   isn't confused by the absence of punctuation.

    my $html_title = htmlquote($title);
    my $html_author = htmlquote($author);
    my $html_isbn = htmlquote($sisbn);

    my $html = '';
    if ($html_title ne '') {
        $html .= "<tr><th align=\"right\">Title: &nbsp;</th> <td>$html_title</td></tr>\n";
    }
    if ($html_author ne '') {
        $html .= "<tr><th align=\"right\">Author: &nbsp;</th> <td>$html_author</td></tr>\n";
    }
    if ($html_isbn ne '') {
        $html .= "<tr><th align=\"right\">ISBN: &nbsp;</th> <td>$html_isbn</td></tr>\n";
    }

    print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<title>Thank you for your book recommendation</title>
<style type="text/css">
    body {
        background-color: #FFFFFF;
        margin-left: 15%;
        margin-right: 10%
    }

    h1 {
        text-align: center;
    }
</style>
<meta name="robots" content="noindex" />
</head>

<body>

<h1>Book Recommendation Received!</h1>

<p>
Thank you for recommending:
</p>

<table>
$html
</table>

<p>
I'll check it out!
</p>
EOF

    #   Generate back link to originating page.  Note that the
    #   $referer URL is already quoted, encoding any impermissible
    #   characters as %XX escapes.

    if ($referer ne '') {
        print("<h4><a href=\"$referer\">$backlink</a></h4>");
    }

    print << "EOF";

EOF

    if ($debug) {
        print << "EOF";
<pre>
$brec
Email: "$email"
Mailer: "$mailer"
EOF

    $a = `sh -c set`;
    print $a;

    print << "EOF";
</pre>
EOF
    }

    print << "EOF";
</body>
</html>
EOF

    close(STDOUT);

    #   Trim string by removing leading and trailing white space

    sub trim {
        my ($s) = @_;

        $s =~ s/^\s+//;         # Delete leading blanks
        $s =~ s/\s+$//;         # Delete trailing blanks
        return $s;
    }

    #   Quote user-supplied text for inclusion in log file

    sub iquote {
        my ($s) = @_;

        $s =~ s/"/""/g;         # Force quote marks with double quotes
        return "\"$s\"";        # Return quoted string
    }

    #   Quote string for inclusion in an HTML document

    sub htmlquote {
        my ($s) = @_;

        $s =~ s/&/&amp;/g;
        $s =~ s/</&lt;/g;
        $s =~ s/>/&gt;/g;
        return $s;
    }

    #   Test whether the argument is a valid ISBN-13.  Punctuation
    #   is ignored.  The function returns the canonical 13 character
    #   ISBN-13 if valid and the null string otherwise.

    sub validate_ISBN_13 {
        my ($s) = @_;
        my (@digits, $sum, $d, $i, $f);

        $s =~ tr/0-9//cd;       # Delete any non-digits
        if (length($s) != 13) {
            return '';          # All valid ISBN-13s are thirteen characters
        }
        @digits = unpack('C13', $s);
        $sum = 0;
        $f = 1;
        for ($i = 0; $i < 13; $i ++) {
            $d = $digits[$i] - ord('0');
            $sum += $f * $d;
            $f = ($f + 2) & 3;
        }
        $sum %= 10;
        if ($sum != 0) {
            return '';          # ISBN-13 checksum is incorrect
        }
        return $s;
    }

    #   Test whether the argument is a valid ISBN.  Punctuation
    #   is ignored and the "X" ISBN-10 checksum may be either upper
    #   or lower case.  The function returns the canonical 10 or 13
    #   character ISBN if valid and the null string otherwise.

    sub validate_ISBN {
        my ($s) = @_;
        my (@digits, $sum, $d, $i);

        $s =~ tr/a-z/A-Z/;      # Actually, just for "X"
        $s =~ tr/0-9A-Z//cd;    # Delete any punctuation
        if (length($s) == 13) {
            return validate_ISBN_13($s);
        }
        if (length($s) != 10) {
            return '';          # All valid ISBNs are ten characters
        }
        if ($s =~ m/A-WYZ/) {
            return '';          # "X" is the only letter permitted in an ISBN
        }
        if (substr($s, 0, 9) =~ m/X/) {
            return '';          # "X" may only appear as the last, checksum, character
        }

        $s =~ m/(\w)$/;
        my $cks = $1;
        if ($cks ne ISBN_10_check($s)) {
            return '';          # ISBN checksum is incorrect
        }

        return $s;
    }

    #   Compute ISBN-10 check digit from a canonical ISBN-10.  Only
    #   the first 9 characters of if the ISBN are examined.

    sub ISBN_10_check {
        my ($s) = @_;
        my (@digits, $sum, $d, $i);

        @digits = unpack('C9', $s);
        $sum = 0;
        for ($i = 0; $i < 9; $i++) {
            $d = $digits[$i] - ord('0');
            $sum += (10 - $i) * $d;
        }
        $sum %= 11;
        $sum = 11 - $sum;
        return ($sum == 10) ? 'X' :
            (($sum == 11) ? 0 : $sum);
    }

    #   Convert an ISBN-10 or ISBN-13 to an ASIN which can be used in
    #   an Amazon "ISBN" query.  The argument is assumed to be in
    #   canonical form with all delimiters removed.

    sub ISBNtoASIN {
        my ($n) = @_;

        my $esbn10;
        if (length($isbn) == 13) {
            if (substr($n, 2, 1) eq '9') {
                $esbn10 = '';
            } else {
                my $isbn10 = substr($n, 3, 9);
                my $ckd = ISBN_10_check($isbn10);
                $isbn10 .= $ckd;
                $esbn10 = $n;
                $esbn10 =~ s/^9\D?7\D?[89]\D?//;
                $esbn10 =~ s/\d(\D*)$/$ckd$1/;
            }
        } else {
            $esbn10 = $n;
        }
        return $esbn10;
    }

    #   Parse the CGI QUERY_STRING argument (placed in the environment when
    #   the Web server launches this script).  This function sets environment
    #   variables named WWW_varname for each field in the form with value the
    #   unquoted contents of each field.  This rather odd way of returning
    #   results ("Why not just use a regular hash?") is used to maintain
    #   compatibility with the UNCGI program.  A Perl script which previously
    #   required execution under the control of UNCGI may be run stand-alone
    #   by calling this function at the top.  This halves the number of processes
    #   the server must fork in order to process the request, and is one
    #   fewer thing for the beleaguered system administrator to install.

    sub parse_query_string {
        my ($rmeth, @args, $instream, $var, $varname, $value);

        $rmeth = $ENV{'REQUEST_METHOD'};    # Request method ("GET" or "POST")
        $rmeth =~ tr/a-z/A-Z/;              # Just in case server doesn't shout
        if ($rmeth eq 'GET') {
            @args = split(/&/, $ENV{'QUERY_STRING'});
        } elsif ($rmeth eq 'POST') {
            read(STDIN, $instream, $ENV{'CONTENT_LENGTH'});
            @args = split(/&/, $instream);
        }
        foreach $var (@args) {
            ($varname, $value) = split(/=/, $var);
            $varname = unescape_URL($varname);
            $value = unescape_URL($value);
            #   Elide any server-side includes a mischevious user may have
            #   embedded in a value string.  This protects against mayhem if
            #   the value is included in an HTML reply.
            $value =~ s/<!--(.|\n)*-->//g;

            #   Define an "uncgi" compatible environment variable for
            #   each name/value pair.

            $ENV{"WWW_$varname"} = $value;
        }
    }

    #   Escape a URL

    sub escape_URL {
        my ($s) = @_;

        $s =~ s/([^A-Za-z0-9\$\-_\.\+\!\*'\(\)])/sprintf("%%%02X", ord($1))/eg;
        return $s;
    }

    #   Unescape a URL-encoded string

    sub unescape_URL {
        my ($s) = @_;

        $s =~ tr/+/ /;
        $s =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

        return $s;
    }

    #   Encode/decode a string in ROT13 encoding

    sub rot13 {
        my ($s) = @_;

        $s =~ tr/A-Za-z/N-ZA-Mn-za-m/;
        return $s;
    }
